type 'a table =
  Empty |
  Branch2 of 'a table * (int * 'a) * 'a table |
  Branch3 of 'a table * (int * 'a) * 'a table * (int * 'a) * 'a table;;

exception Dup of int;;

let empty = Empty;;

let find key tab =
  let rec look = function
        Empty -> raise Not_found
      | (Branch2 (left, (k, x), right)) ->
          (match compare key k with
            -1 -> look left
          | 0 -> x
          | 1 -> look right)
      | (Branch3 (left, (k1, x1), mid, (k2, x2), right)) ->
          (match compare key k1 with
            -1 -> look left
          | 0 -> x1
          | 1 ->
              (match compare key k2 with
                -1 -> look mid
              | 0 -> x2
              | 1 -> look right)) in
  look tab;;

let mem tab key =
  let rec def = function
        Empty -> false
      | (Branch2 (left, (k, x), right)) ->
          (match compare key k with
            -1 -> def left
          | 0 -> true
          | 1 -> def right)
      | (Branch3 (left, (k1, x1), mid, (k2, x2), right)) ->
          (match compare key k1 with
            -1 -> def left
          | 0 -> true
          | 1 ->
              (match compare key k2 with
                -1 -> def mid
              | 0 -> true
              | 1 -> def right)) in
  def tab;;

type 'a growth =
  Stay of 'a table |
  Sprout of 'a table * (int * 'a) * 'a table;;

exception Same;;

let rec modify key f tab =
  let rec modfy = function
        Empty -> Sprout (Empty, (key, f None), Empty)
      | (Branch2 (left, ((k, x) as p), right)) ->
          (match compare key k with
            -1 ->
              (match modfy left with
                Stay left' -> Stay (Branch2 (left', p, right))
              | Sprout (left1, q, left2) -> Stay (Branch3 (left1, q, left2, p, right)))
          | 0 -> Stay (Branch2 (left, (k, f (Some x)), right))
          | 1 ->
              (match modfy right with
                Stay right' -> Stay (Branch2 (left, p, right'))
              | Sprout (right1, q, right2) ->
                  Stay (Branch3 (left, p, right1, q, right2))))
      | (Branch3 (left, ((k1, x1) as p1), mid, ((k2, x2) as p2), right)) ->
          (match compare key k1 with
            -1 ->
              (match modfy left with
                Stay left' -> Stay (Branch3 (left', p1, mid, p2, right))
              | Sprout (left1, q, left2) ->
                  Sprout (Branch2 (left1, q, left2), p1, Branch2 (mid, p2, right)))
          | 0 -> Stay (Branch3 (left, (k1, f (Some x1)), mid, p2, right))
          | 1 ->
              (match compare key k2 with
                -1 ->
                  (match modfy mid with
                    Stay mid' -> Stay (Branch3 (left, p1, mid', p2, right))
                  | Sprout (mid1, q, mid2) ->
                      Sprout (Branch2 (left, p1, mid1), q, Branch2 (mid2, p2, right)))
              | 0 -> Stay (Branch3 (left, p1, mid, (k2, f (Some x2)), right))
              | 1 ->
                  (match modfy right with
                    Stay right' -> Stay (Branch3 (left, p1, mid, p2, right'))
                  | Sprout (right1, q, right2) ->
                      Sprout (Branch2 (left, p1, mid), p2, Branch2 (right1, q, right2))))) in

  try match modfy tab with
      Stay tab' -> tab'
    | Sprout (a,b,c) -> Branch2 (a,b,c)
  with Same -> tab
;;

let add key x = modify key (fun _ -> x);;
